/*
 * Copyright (c) 2003 Sun Microsystems, Inc. All Rights Reserved.
 * Copyright (c) 2010 JogAmp Community. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * - Redistribution of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES,
 * INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN
 * MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR
 * ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR
 * DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE
 * DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY,
 * ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF
 * SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * You acknowledge that this software is not designed or intended for use
 * in the design, construction, operation or maintenance of any nuclear
 * facility.
 * 
 * Sun gratefully acknowledges that this software was originally authored
 * and developed by Kenneth Bradley Russell and Christopher John Kline.
 */

package javax.media.opengl;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.List;

import com.jogamp.common.JogampRuntimeException;

import com.jogamp.common.util.ReflectionUtil;
import com.jogamp.opengl.GLAutoDrawableDelegate;
import com.jogamp.opengl.GLRendererQuirks;

import javax.media.nativewindow.AbstractGraphicsDevice;
import javax.media.nativewindow.AbstractGraphicsScreen;
import javax.media.nativewindow.AbstractGraphicsConfiguration;
import javax.media.nativewindow.CapabilitiesImmutable;
import javax.media.nativewindow.NativeSurface;
import javax.media.nativewindow.NativeWindowFactory;
import javax.media.nativewindow.ProxySurface;
import javax.media.nativewindow.UpstreamSurfaceHook;

import jogamp.opengl.Debug;

/** <P> Provides a virtual machine- and operating system-independent
    mechanism for creating {@link GLDrawable}s. </P>

    <P> The {@link javax.media.opengl.GLCapabilities} objects passed
    in to the various factory methods are used as a hint for the
    properties of the returned drawable. The default capabilities
    selection algorithm (equivalent to passing in a null {@link
    GLCapabilitiesChooser}) is described in {@link
    DefaultGLCapabilitiesChooser}. Sophisticated applications needing
    to change the selection algorithm may pass in their own {@link
    GLCapabilitiesChooser} which can select from the available pixel
    formats. The GLCapabilitiesChooser mechanism may not be supported
    by all implementations or on all platforms, in which case any
    passed GLCapabilitiesChooser will be ignored. </P>

    <P> Because of the multithreaded nature of the Java platform's
    Abstract Window Toolkit, it is typically not possible to immediately
    reject a given {@link GLCapabilities} as being unsupportable by
    either returning <code>null</code> from the creation routines or
    raising a {@link GLException}. The semantics of the rejection
    process are (unfortunately) left unspecified for now. The current
    implementation will cause a {@link GLException} to be raised
    during the first repaint of the {@link javax.media.opengl.awt.GLCanvas} or {@link
    javax.media.opengl.awt.GLJPanel} if the capabilities can not be met.<br>
    {@link javax.media.opengl.GLPbuffer} are always
    created immediately and their creation will fail with a 
    {@link javax.media.opengl.GLException} if errors occur. </P>

    <P> The concrete GLDrawableFactory subclass instantiated by {@link
    #getFactory getFactory} can be changed by setting the system
    property <code>opengl.factory.class.name</code> to the
    fully-qualified name of the desired class. </P>
*/
public abstract class GLDrawableFactory {

  protected static final boolean DEBUG = Debug.debug("GLDrawable");
  
  /** 
   * We have to disable support for ANGLE, the D3D ES2 emulation on Windows provided w/ Firefox and Chrome. 
   * When run in the mentioned browsers, the eglInitialize(..) implementation crashes.
   * <p>
   * This can be overridden by explicitly enabling ANGLE on Windows by setting the property
   * <code>jogl.enable.ANGLE</code>.
   * </p> 
   */
  protected static final boolean enableANGLE = Debug.isPropertyDefined("jogl.enable.ANGLE", true);

  /** 
   * In case no OpenGL ES implementation is required
   * and if the running platform may have a buggy implementation,
   * setting the property <code>jogl.disable.opengles</code> disables querying a possible existing OpenGL ES implementation. 
   */
  protected static final boolean disableOpenGLES = Debug.isPropertyDefined("jogl.disable.opengles", true);

  static final String macosxFactoryClassNameCGL = "jogamp.opengl.macosx.cgl.MacOSXCGLDrawableFactory";
  static final String macosxFactoryClassNameAWTCGL = "jogamp.opengl.macosx.cgl.awt.MacOSXAWTCGLDrawableFactory";

  private static volatile boolean isInit = false;
  private static GLDrawableFactory eglFactory;
  private static GLDrawableFactory nativeOSFactory;

  protected static ArrayList<GLDrawableFactory> glDrawableFactories = new ArrayList<GLDrawableFactory>();

  // Shutdown hook mechanism for the factory
  private static boolean factoryShutdownHookRegistered = false;
  private static Thread factoryShutdownHook = null;
  private static volatile boolean isJVMShuttingDown = false;

  /**
   * Instantiate singleton factories if available, EGLES1, EGLES2 and the OS native ones.
   */
  public static final void initSingleton() { 
      if (!isInit) { // volatile: ok
          synchronized (GLDrawableFactory.class) {
              if (!isInit) {
                  isInit=true;
                  initSingletonImpl();
              }
          }
      }
  }  
  private static final void initSingletonImpl() {
    registerFactoryShutdownHook();
    
    final String nwt = NativeWindowFactory.getNativeWindowType(true);
    GLDrawableFactory tmp = null;
    String factoryClassName = Debug.getProperty("jogl.gldrawablefactory.class.name", true);
    ClassLoader cl = GLDrawableFactory.class.getClassLoader();
    if (null == factoryClassName) {
        if ( nwt == NativeWindowFactory.TYPE_X11 ) {
          factoryClassName = "jogamp.opengl.x11.glx.X11GLXDrawableFactory";
        } else if ( nwt == NativeWindowFactory.TYPE_WINDOWS ) {
          factoryClassName = "jogamp.opengl.windows.wgl.WindowsWGLDrawableFactory";
        } else if ( nwt == NativeWindowFactory.TYPE_MACOSX ) {
            if(ReflectionUtil.isClassAvailable(macosxFactoryClassNameAWTCGL, cl)) {
                factoryClassName = macosxFactoryClassNameAWTCGL;
            } else {
                factoryClassName = macosxFactoryClassNameCGL;
            }
        } else {
          // may use egl*Factory ..
          if (DEBUG || GLProfile.DEBUG) {
              System.err.println("GLDrawableFactory.static - No native Windowing Factory for: "+nwt+"; May use EGLDrawableFactory, if available." );
          }
        }
    }
    if (null != factoryClassName) {
      if (DEBUG || GLProfile.DEBUG) {
          System.err.println("GLDrawableFactory.static - Native OS Factory for: "+nwt+": "+factoryClassName);
      }
      try {
          tmp = (GLDrawableFactory) ReflectionUtil.createInstance(factoryClassName, cl);
      } catch (JogampRuntimeException jre) { 
          if (DEBUG || GLProfile.DEBUG) {
              System.err.println("Info: GLDrawableFactory.static - Native Platform: "+nwt+" - not available: "+factoryClassName);
              jre.printStackTrace();
          }
      }
    }
    if(null != tmp && tmp.isComplete()) {
        nativeOSFactory = tmp;
    }
    tmp = null;

    if(!disableOpenGLES) {
        try {
            tmp = (GLDrawableFactory) ReflectionUtil.createInstance("jogamp.opengl.egl.EGLDrawableFactory", cl);
        } catch (JogampRuntimeException jre) {
            if (DEBUG || GLProfile.DEBUG) {
                System.err.println("Info: GLDrawableFactory.static - EGLDrawableFactory - not available");
                jre.printStackTrace();
            }
        }
        if(null != tmp && tmp.isComplete()) {
            eglFactory = tmp;
        }
    } else if( DEBUG || GLProfile.DEBUG ) {
        System.err.println("Info: GLDrawableFactory.static - EGLDrawableFactory - disabled!");
    }
  }

  protected static void shutdown() {
    if (isInit) { // volatile: ok
      synchronized (GLDrawableFactory.class) {
          if (isInit) {
              isInit=false;
              shutdownImpl();
          }
      }
    }
  }
  
  private static void shutdownImpl() {
    // Following code will _always_ remain in shutdown hook
    // due to special semantics of native utils, i.e. X11Utils.
    // The latter requires shutdown at JVM-Shutdown only.
    synchronized(glDrawableFactories) {
        for(int i=0; i<glDrawableFactories.size(); i++) {
            glDrawableFactories.get(i).destroy();
        }
        glDrawableFactories.clear();
        
        // both were members of glDrawableFactories and are shutdown already 
        nativeOSFactory = null;
        eglFactory = null;
    }
    GLContext.shutdown();
    NativeWindowFactory.shutdown(isJVMShuttingDown);
  }
  
  private static synchronized void registerFactoryShutdownHook() {
    if (factoryShutdownHookRegistered) {
        return;
    }
    factoryShutdownHook = new Thread(new Runnable() {
        public void run() {
            isJVMShuttingDown = true;
            GLDrawableFactory.shutdownImpl();
        }
    });
    AccessController.doPrivileged(new PrivilegedAction<Object>() {
        public Object run() {
            Runtime.getRuntime().addShutdownHook(factoryShutdownHook);
            return null;
        }
    });
    factoryShutdownHookRegistered = true;
  }

  protected GLDrawableFactory() {
    synchronized(glDrawableFactories) {
        glDrawableFactories.add(this);
    }
  }
  
  /** Returns true if this factory is complete, i.e. ready to be used. Otherwise return false. */
  protected abstract boolean isComplete();

  protected void enterThreadCriticalZone() {};
  protected void leaveThreadCriticalZone() {};

  protected abstract void destroy();

  /**
   * Retrieve the default <code>device</code> {@link AbstractGraphicsDevice#getConnection() connection},
   * {@link AbstractGraphicsDevice#getUnitID() unit ID} and {@link AbstractGraphicsDevice#getUniqueID() unique ID name}. for this factory<br>
   * The implementation must return a non <code>null</code> default device, which must not be opened, ie. it's native handle is <code>null</code>.
   * <p>
   * This method shall return the default device if available
   * even if the GLDrawableFactory is not functional and hence not compatible. 
   * The latter situation may happen because no native OpenGL implementation is available for the specific implementation.
   * </p>
   * @return the default shared device for this factory, eg. :0.0 on X11 desktop.
   * @see #getIsDeviceCompatible(AbstractGraphicsDevice)
   */
  public abstract AbstractGraphicsDevice getDefaultDevice();

  /**
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared the target device, may be <code>null</code> for the platform's default device.
   * @return true if the device is compatible with this factory, ie. if it can be used for GLDrawable creation. Otherwise false.
   *         This implies validation whether the implementation is functional.
   *         
   * @see #getDefaultDevice()
   */
  public abstract boolean getIsDeviceCompatible(AbstractGraphicsDevice device);

  protected final AbstractGraphicsDevice validateDevice(AbstractGraphicsDevice device) {
      if(null==device) {
          device = getDefaultDevice();
          if(null==device) {
              throw new InternalError("no default device");
          }
          if (GLProfile.DEBUG) {
              System.err.println("Info: "+getClass().getSimpleName()+".validateDevice: using default device : "+device);
          }
      }
      
      // Always validate the device, 
      // since even the default device may not be used by this factory.
      if( !getIsDeviceCompatible(device) ) {
          if (GLProfile.DEBUG) {
              System.err.println("Info: "+getClass().getSimpleName()+".validateDevice: device not compatible : "+device);
          }
          return null;
      }
      return device;
  }

  /**
   * Validate and start the shared resource runner thread if necessary and 
   * if the implementation uses it.
   * 
   * @return the shared resource runner thread, if implementation uses it.
   */
  protected abstract Thread getSharedResourceThread();
  
  /**
   * Create the shared resource used internally as a reference for capabilities etc.
   * <p>
   * Returns true if a shared resource could be created 
   * for the <code>device</code> {@link AbstractGraphicsDevice#getConnection()}.<br>
   * This does not imply a shared resource is mapped (ie. made persistent), but is available in general<br>.
   * </p>
   *
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared the target device, may be <code>null</code> for the platform's default device.
   * @return true if a shared resource could been created, otherwise false. 
   */
  protected abstract boolean createSharedResource(AbstractGraphicsDevice device);
  
  /**
   * Returns true if the <code>quirk</code> exist in the shared resource's context {@link GLRendererQuirks}.
   * <p>
   * Convenience method for:
   * <pre>
      final GLRendererQuirks glrq = factory.getRendererQuirks(device);
      return null != glrq ? glrq.exist(quirk) : false;
   * </pre>
   * </p>
   * 
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared the target device, may be <code>null</code> for the platform's default device.
   * @param quirk the quirk to be tested, e.g. {@link GLRendererQuirks#NoDoubleBufferedPBuffer}.
   * @throws IllegalArgumentException if the quirk is out of range
   * @see #getRendererQuirks()
   * @see GLRendererQuirks
   */
  public final boolean hasRendererQuirk(AbstractGraphicsDevice device, int quirk) {
      final GLRendererQuirks glrq = getRendererQuirks(device);
      return null != glrq ? glrq.exist(quirk) : false;
  }
  
  /**
   * Returns the shared resource's context {@link GLRendererQuirks}.
   * <p>
   * Implementation calls {@link GLContext#getRendererQuirks()} on the shared resource context.
   * </p>
   * <p>
   * In case no shared device exist yet or the implementation doesn't support tracking quirks,
   * the result is always <code>null</code>.
   * </p>
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared the target device, may be <code>null</code> for the platform's default device.
   * @see GLContext#getRendererQuirks()
   * @see GLRendererQuirks
   */
  public abstract GLRendererQuirks getRendererQuirks(AbstractGraphicsDevice device);
  
  /**
   * Returns the sole GLDrawableFactory instance for the desktop (X11, WGL, ..) if exist or null
   */
  public static GLDrawableFactory getDesktopFactory() {
    GLProfile.initSingleton();    
    return nativeOSFactory;
  }

  /**
   * Returns the sole GLDrawableFactory instance for EGL if exist or null
   */
  public static GLDrawableFactory getEGLFactory() {
    GLProfile.initSingleton();    
    return eglFactory;
  }

  /** 
   * Returns the sole GLDrawableFactory instance. 
   * 
   * @param glProfile GLProfile to determine the factory type, ie EGLDrawableFactory, 
   *                or one of the native GLDrawableFactory's, ie X11/GLX, Windows/WGL or MacOSX/CGL.
   */
  public static GLDrawableFactory getFactory(GLProfile glProfile) throws GLException {
    return getFactoryImpl(glProfile.getImplName());
  }

  protected static GLDrawableFactory getFactoryImpl(String glProfileImplName) throws GLException {
    if ( GLProfile.usesNativeGLES(glProfileImplName) ) {
        if(null!=eglFactory) {   
            return eglFactory;
        }
    } else if(null!=nativeOSFactory) {
        return nativeOSFactory;
    }
    throw new GLException("No GLDrawableFactory available for profile: "+glProfileImplName);
  }

  protected static GLDrawableFactory getFactoryImpl(AbstractGraphicsDevice device) throws GLException {
    if(null != nativeOSFactory && nativeOSFactory.getIsDeviceCompatible(device)) {
        return nativeOSFactory;
    }
    if(null != eglFactory && eglFactory.getIsDeviceCompatible(device)) {
        return eglFactory;
    }
    throw new GLException("No native platform GLDrawableFactory, nor EGLDrawableFactory available: "+device);
  }

  /**
   * Returns an array of available GLCapabilities for the device.<br>
   * The list is sorted by the native ID, ascending.<br>
   * The chosen GLProfile statement in the result may not refer to the maximum available profile
   * due to implementation constraints, ie using the shared resource.
   *
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared the target device, may be <code>null</code> for the platform's default device.
   * @return A list of {@link javax.media.opengl.GLCapabilitiesImmutable}'s, maybe empty if none is available.
   */
  public final List<GLCapabilitiesImmutable> getAvailableCapabilities(AbstractGraphicsDevice device) {
      device = validateDevice(device);
      if(null!=device) {
        return getAvailableCapabilitiesImpl(device);
      }
      return null;
  }
  protected abstract List<GLCapabilitiesImmutable> getAvailableCapabilitiesImpl(AbstractGraphicsDevice device);

  //----------------------------------------------------------------------
  // Methods to create high-level objects

  /**
   * Returns a GLDrawable according to it's chosen {@link GLCapabilitiesImmutable},<br>
   * which determines pixel format, on- and offscreen incl. PBuffer type.
   * <p>
   * The chosen {@link GLCapabilitiesImmutable} are referenced within the target
   * {@link NativeSurface}'s {@link AbstractGraphicsConfiguration}.<p>
   * </p>
   * <p>
   * An onscreen GLDrawable is created if {@link CapabilitiesImmutable#isOnscreen() caps.isOnscreen()} is true.
   * </p>
   * <p>
   * A FBO drawable is created if both {@link GLCapabilitiesImmutable#isFBO() caps.isFBO()}
   * and {@link GLContext#isFBOAvailable(AbstractGraphicsDevice, GLProfile) canCreateFBO(device, caps.getGLProfile())} is true.
   * </p>
   * <p>
   * A Pbuffer drawable is created if both {@link GLCapabilitiesImmutable#isPBuffer() caps.isPBuffer()}
   * and {@link #canCreateGLPbuffer(AbstractGraphicsDevice) canCreateGLPbuffer(device)} is true.
   * </p>
   * <p>
   * If not onscreen and neither FBO nor Pbuffer is available, 
   * a simple pixmap/bitmap drawable/surface is created, which is unlikely to be hardware accelerated.
   * </p>
   * 
   * @throws IllegalArgumentException if the passed target is null
   * @throws GLException if any window system-specific errors caused
   *         the creation of the GLDrawable to fail.
   *
   * @see #canCreateGLPbuffer(AbstractGraphicsDevice)
   * @see GLContext#isFBOAvailable(AbstractGraphicsDevice, GLProfile)
   * @see javax.media.opengl.GLCapabilities#isOnscreen()
   * @see javax.media.opengl.GLCapabilities#isFBO()
   * @see javax.media.opengl.GLCapabilities#isPBuffer()
   * @see javax.media.nativewindow.GraphicsConfigurationFactory#chooseGraphicsConfiguration(Capabilities, CapabilitiesChooser, AbstractGraphicsScreen)
   */
  public abstract GLDrawable createGLDrawable(NativeSurface target)
    throws IllegalArgumentException, GLException;
  
  /**
   * Creates an {@link GLOffscreenAutoDrawable} incl it's offscreen {@link javax.media.nativewindow.NativeSurface} with the given capabilites and dimensions.
   * <p>
   * The {@link GLOffscreenAutoDrawable}'s {@link GLDrawable} is realized and it's {@link GLContext} assigned but not yet made current.
   * </p>
   * <p>
   * In case the passed {@link GLCapabilitiesImmutable} contains default values, i.e.
   * {@link GLCapabilitiesImmutable#isOnscreen() caps.isOnscreen()} <code> == true</code>,
   * it is auto-configured. The latter will set offscreen and also FBO <i>or</i> Pbuffer, whichever is available in that order.
   * </p>
   * <p>
   * A FBO based auto drawable, {@link GLOffscreenAutoDrawable.FBO}, is created if both {@link GLCapabilitiesImmutable#isFBO() caps.isFBO()}
   * and {@link GLContext#isFBOAvailable(AbstractGraphicsDevice, GLProfile) canCreateFBO(device, caps.getGLProfile())} is true.
   * </p>
   * <p>
   * A Pbuffer based auto drawable is created if both {@link GLCapabilitiesImmutable#isPBuffer() caps.isPBuffer()}
   * and {@link #canCreateGLPbuffer(AbstractGraphicsDevice) canCreateGLPbuffer(device)} is true.
   * </p>
   * <p>
   * If neither FBO nor Pbuffer is available, 
   * a simple pixmap/bitmap auto drawable is created, which is unlikely to be hardware accelerated.
   * </p>
   *
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared device to be used, may be <code>null</code> for the platform's default device.
   * @param caps the requested GLCapabilties
   * @param chooser the custom chooser, may be null for default
   * @param width the requested offscreen width
   * @param height the requested offscreen height
   * @return the created and initialized offscreen {@link GLOffscreenAutoDrawable} instance
   *
   * @throws GLException if any window system-specific errors caused
   *         the creation of the Offscreen to fail.
   *         
   * @see #createOffscreenDrawable(AbstractGraphicsDevice, GLCapabilitiesImmutable, GLCapabilitiesChooser, int, int)
   */
  public abstract GLOffscreenAutoDrawable createOffscreenAutoDrawable(AbstractGraphicsDevice device,
                                                                      GLCapabilitiesImmutable caps,
                                                                      GLCapabilitiesChooser chooser,
                                                                      int width, int height,
                                                                      GLContext shareWith) throws GLException;
  /**
   * Creates a offscreen {@link GLDrawable} incl it's offscreen {@link javax.media.nativewindow.NativeSurface} with the given capabilites and dimensions.
   * <p>
   * In case the passed {@link GLCapabilitiesImmutable} contains default values, i.e.
   * {@link GLCapabilitiesImmutable#isOnscreen() caps.isOnscreen()} <code> == true</code>,
   * it is auto-configured. The latter will set offscreen and also FBO <i>or</i> Pbuffer, whichever is available in that order.
   * </p>
   * <p>
   * A resizeable FBO drawable, {@link GLFBODrawable.Resizeable}, is created if both {@link GLCapabilitiesImmutable#isFBO() caps.isFBO()}
   * and {@link GLContext#isFBOAvailable(AbstractGraphicsDevice, GLProfile) canCreateFBO(device, caps.getGLProfile())} is true.
   * </p>
   * <p>
   * A Pbuffer drawable is created if both {@link GLCapabilitiesImmutable#isPBuffer() caps.isPBuffer()}
   * and {@link #canCreateGLPbuffer(AbstractGraphicsDevice) canCreateGLPbuffer(device)} is true.
   * </p>
   * <p>
   * If neither FBO nor Pbuffer is available, 
   * a simple pixmap/bitmap drawable is created, which is unlikely to be hardware accelerated.
   * </p>
   *
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared device to be used, may be <code>null</code> for the platform's default device.
   * @param caps the requested GLCapabilties
   * @param chooser the custom chooser, may be null for default
   * @param width the requested offscreen width
   * @param height the requested offscreen height
   *
   * @return the created offscreen {@link GLDrawable}
   *
   * @throws GLException if any window system-specific errors caused
   *         the creation of the Offscreen to fail.
   *         
   * @see #createOffscreenAutoDrawable(AbstractGraphicsDevice, GLCapabilitiesImmutable, GLCapabilitiesChooser, int, int, GLContext)
   */
  public abstract GLDrawable createOffscreenDrawable(AbstractGraphicsDevice device,
                                                     GLCapabilitiesImmutable caps,
                                                     GLCapabilitiesChooser chooser,
                                                     int width, int height) throws GLException;

  /**
   * Creates a proxy {@link NativeSurface} w/ defined surface handle, i.e. a {@link WrappedSurface} or {@link GDISurface} instance. 
   * <p>
   * It's {@link AbstractGraphicsConfiguration} is properly set according to the given 
   * <code>windowHandle</code>'s native visualID if set or the given {@link GLCapabilitiesImmutable}.
   * </p>
   * <p>
   * Lifecycle (destruction) of the given surface handle shall be handled by the caller.
   * </p>
   * <p>
   * Such surface can be used to instantiate a GLDrawable. With the help of {@link GLAutoDrawableDelegate}
   * you will be able to implement a new native windowing system  binding almost on-the-fly, see {@link com.jogamp.opengl.swt.GLCanvas}. 
   * </p>
   * 
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared the target device, may be <code>null</code> for the platform's default device.
   *        Caller has to ensure it is compatible w/ the given <code>windowHandle</code> 
   * @param screenIdx matching screen index of given <code>windowHandle</code>
   * @param windowHandle the native window handle
   * @param caps the requested GLCapabilties
   * @param chooser the custom chooser, may be null for default
   * @param upstream optional {@link ProxySurface.UpstreamSurfaceHook} allowing control of the {@link ProxySurface}'s lifecycle and data it presents.
   * @return the created {@link ProxySurface} instance w/ defined surface handle.
   */
  public abstract ProxySurface createProxySurface(AbstractGraphicsDevice device,
                                                  int screenIdx, 
                                                  long windowHandle, 
                                                  GLCapabilitiesImmutable caps, GLCapabilitiesChooser chooser, UpstreamSurfaceHook upstream);

  /**
   * Returns true if it is possible to create an <i>framebuffer object</i> (FBO).
   * <p>
   * FBO feature is implemented in OpenGL, hence it is {@link GLProfile} dependent.
   * </p> 
   * <p>
   * FBO support is queried as described in {@link GLContext#hasBasicFBOSupport()}.
   * </p>
   *
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared the target device, may be <code>null</code> for the platform's default device.
   * @param glp {@link GLProfile} to check for FBO capabilities
   * @see GLContext#hasBasicFBOSupport()
   */
  public abstract boolean canCreateFBO(AbstractGraphicsDevice device, GLProfile glp);

  /**
   * Returns true if it is possible to create a GLPbuffer. Some older
   * graphics cards do not have this capability.
   *
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared the target device, may be <code>null</code> for the platform's default device.
   */
  public abstract boolean canCreateGLPbuffer(AbstractGraphicsDevice device);

  /**
   * Creates a GLPbuffer {@link GLAutoDrawable} with the given capabilites and dimensions.
   * <p>
   * The GLPbuffer drawable is realized and initialized eagerly.
   * </p>
   *
   * See the note in the overview documentation on
   * <a href="../../../overview-summary.html#SHARING">context sharing</a>.
   *
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared the target device, may be <code>null</code> for the platform's default device.
   * @param capabilities the requested capabilities
   * @param chooser the custom chooser, may be null for default
   * @param initialWidth initial width of pbuffer
   * @param initialHeight initial height of pbuffer
   * @param shareWith a shared GLContext this GLPbuffer shall use
   *
   * @return the created and initialized {@link GLPbuffer} instance
   *
   * @throws GLException if any window system-specific errors caused
   *         the creation of the GLPbuffer to fail.
   *         
   * @deprecated {@link GLPbuffer} is deprecated, use {@link #createOffscreenAutoDrawable(AbstractGraphicsDevice, GLCapabilitiesImmutable, GLCapabilitiesChooser, int, int, GLContext)}
   */
  public abstract GLPbuffer createGLPbuffer(AbstractGraphicsDevice device,
                                            GLCapabilitiesImmutable capabilities,
                                            GLCapabilitiesChooser chooser,
                                            int initialWidth,
                                            int initialHeight,
                                            GLContext shareWith)
    throws GLException;

  
  //----------------------------------------------------------------------
  // Methods for interacting with third-party OpenGL libraries

  /**
   * <P> Creates a GLContext object representing an existing OpenGL
   * context in an external (third-party) OpenGL-based library. This
   * GLContext object may be used to draw into this preexisting
   * context using its {@link GL} and {@link
   * javax.media.opengl.glu.GLU} objects. New contexts created through
   * {@link GLDrawable}s may share textures and display lists with
   * this external context. </P>
   *
   * <P> The underlying OpenGL context must be current on the current
   * thread at the time this method is called. The user is responsible
   * for the maintenance of the underlying OpenGL context; calls to
   * <code>makeCurrent</code> and <code>release</code> on the returned
   * GLContext object have no effect. If the underlying OpenGL context
   * is destroyed, the <code>destroy</code> method should be called on
   * the <code>GLContext</code>. A new <code>GLContext</code> object
   * should be created for each newly-created underlying OpenGL
   * context.
   *
   * @throws GLException if any window system-specific errors caused
   *         the creation of the external GLContext to fail.
   */
  public abstract GLContext createExternalGLContext()
    throws GLException;

  /**
   * Returns true if it is possible to create an external GLDrawable
   * object via {@link #createExternalGLDrawable}.
   *
   * @param device which {@link javax.media.nativewindow.AbstractGraphicsDevice#getConnection() connection} denotes the shared the target device, may be <code>null</code> for the platform's default device.
   */
  public abstract boolean canCreateExternalGLDrawable(AbstractGraphicsDevice device);

  /**
   * <P> Creates a {@link GLDrawable} object representing an existing
   * OpenGL drawable in an external (third-party) OpenGL-based
   * library. This GLDrawable object may be used to create new,
   * fully-functional {@link GLContext}s on the OpenGL drawable. This
   * is useful when interoperating with a third-party OpenGL-based
   * library and it is essential to not perturb the state of the
   * library's existing context, even to the point of not sharing
   * textures or display lists with that context. </P>
   *
   * <P> An underlying OpenGL context must be current on the desired
   * drawable and the current thread at the time this method is
   * called. The user is responsible for the maintenance of the
   * underlying drawable. If one or more contexts are created on the
   * drawable using {@link GLDrawable#createContext}, and the drawable
   * is deleted by the third-party library, the user is responsible
   * for calling {@link GLContext#destroy} on these contexts. </P>
   *
   * <P> Calls to <code>setSize</code>, <code>getWidth</code> and
   * <code>getHeight</code> are illegal on the returned GLDrawable. If
   * these operations are required by the user, they must be performed
   * by the third-party library. </P>
   *
   * <P> It is legal to create both an external GLContext and
   * GLDrawable representing the same third-party OpenGL entities.
   * This can be used, for example, to query current state information
   * using the external GLContext and then create and set up new
   * GLContexts using the external GLDrawable. </P>
   *
   * <P> This functionality may not be available on all platforms and
   * {@link #canCreateExternalGLDrawable} should be called first to
   * see if it is present. For example, on X11 platforms, this API
   * requires the presence of GLX 1.3 or later.
   *
   * @throws GLException if any window system-specific errors caused
   *         the creation of the external GLDrawable to fail.
   */
  public abstract GLDrawable createExternalGLDrawable()
    throws GLException;
}
